from sqlalchemy import String, Boolean, DateTime
from sqlalchemy.schema import Column
from libpkpass.models import Base

class Cert(Base):
    """Table of Certificates"""
    __tablename__ = 'cert'
    cert_bytes = Column(String())
    verified = Column(Boolean)
    fingerprint = Column(String(), primary_key=True)
    subject = Column(String())
    issuer = Column(String())
    enddate = Column(DateTime())
    issuerhash = Column(String())
    subjecthash = Column(String())

        #########################################
    def __repr__(self):
        #########################################
        return f"<Cert {self.fingerprint}>"

    def __iter__(self):
        for c in self.__table__.columns:
            yield c.name, getattr(self, c.name)



#!/usr/bin/env python3
"""This Module tests crypto functionality"""
import unittest
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from libpkpass.crypto import pk_encrypt_string, pk_decrypt_string, \
    pk_sign_string, pk_verify_signature, get_cert_fingerprint
from libpkpass.identities import IdentityDB
from libpkpass.models.recipient import Recipient
from libpkpass.models.cert import Cert

class TestBasicFunction(unittest.TestCase):
    """This class tests the crypto class"""
    def setUp(self):
        self.plaintext = "PLAINTEXT"

        self.certdir = 'test/pki/intermediate/certs'
        self.keydir = 'test/pki/intermediate/private'
        self.cabundle = 'test/pki/intermediate/certs/ca-bundle'

        self.identities = IdentityDB()
        db_path = 'test/pki/intermediate/certs/rd.db'
        self.identities.args = {
            'db': {
                'uri': f"sqlite+pysqlite:///{db_path}",
                'engine': create_engine(f"sqlite+pysqlite:///{db_path}")
            }
        }
        self.session = sessionmaker(bind=self.identities.args['db']['engine'])()
        self.identities.load_certs_from_directory(self.certdir, self.cabundle)
        self.identities.load_keys_from_directory(self.keydir)

    # Encrypt strings for all test identities and make sure they are different
    def test_encrypt_string(self):
        """Test encrypting a string"""
        results = []
        for identity in self.session.query(Recipient).all():
            cert = self.session.query(Cert).filter(
                Cert.recipients.contains(identity)
            ).first()
            results.append(
                pk_encrypt_string(
                    self.plaintext, cert.cert_bytes
                )
            )
        self.assertTrue(results[0] != results[1])

    # Encrypt/Decrypt strings for all test identities and make sure we get back what we put in
        for identity in self.session.query(Recipient).all():
            cert = self.session.query(Cert).filter(
                Cert.recipients.contains(identity)
            ).first()
            (ciphertext, derived_key) = pk_encrypt_string(
                self.plaintext, cert.cert_bytes)
            plaintext = pk_decrypt_string(
                ciphertext, derived_key, dict(identity), None)
            self.assertEqual(self.plaintext, plaintext)

    def test_verify_string(self):
        """verify string is correct"""
        results = []
        for identity in self.session.query(Recipient).all():
            results.append(pk_sign_string(
                self.plaintext, dict(identity), None))
        self.assertTrue(results[0] != results[1])

        for identity in self.session.query(Recipient).all():
            signature = pk_sign_string(self.plaintext, dict(identity), None)
            cert = self.session.query(Cert).filter(
                Cert.recipients.contains(identity)
            ).first()
            self.assertTrue(pk_verify_signature(
                self.plaintext, signature, [cert]))

    def test_cert_fingerprint(self):
        """Verify fingerprint is correct"""
        for identity in self.session.query(Recipient).all():
            cert = self.session.query(Cert).filter(
                Cert.recipients.contains(identity)
            ).first()
            fingerprint = get_cert_fingerprint(cert.cert_bytes)
            self.assertTrue(len(fingerprint.split(':')) == 20)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""This module tests the rename module"""
import builtins
import unittest
import mock
from libpkpass.commands.cli import Cli
from libpkpass.errors import DecryptionError, CliArgumentError
from .basetest.basetest import patch_args, ERROR_MSGS

class RenameTests(unittest.TestCase):
    """This class tests the rename class"""
    def test_recipient_not_in_database(self):
        """Test what happens when a recipient is not in the appropriate directory"""
        with self.assertRaises(CliArgumentError) as context:
            with patch_args(subparser_name='rename', identity='bleh', nopassphrase='true',
                            all=None, pwname='test', rename='retest'):
                Cli().run()
        self.assertEqual(
            context.exception.msg,
            ERROR_MSGS['rep']
        )

    def test_rename_success(self):
        """Test a successful rename (then rename it back)"""
        ret = True
        try:
            with patch_args(subparser_name='rename', identity='r1', nopassphrase='true',
                            all=True, pwname='test', rename='retest', overwrite='true'):
                with mock.patch.object(builtins, 'input', lambda _: 'y'):
                    Cli().run()
            with patch_args(subparser_name='rename', identity='r1', nopassphrase='true',
                            all=True, pwname='retest', rename='test', overwrite="true"):
                with mock.patch.object(builtins, 'input', lambda _: 'y'):
                    Cli().run()
        except DecryptionError:
            ret = False
        self.assertTrue(ret)

    def test_rename_cli_error(self):
        """Test what happens when pwname is not supplied"""
        with self.assertRaises(CliArgumentError) as context:
            with patch_args(subparser_name='rename', identity='r1', nopassphrase='true',
                            all=True, rename='test', overwrite='true'):
                Cli().run()
        self.assertEqual(context.exception.msg, ERROR_MSGS['pwname'])

if __name__ == '__main__':
    unittest.main()



import logging
from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

LOGGER = logging.getLogger(__name__)
logging.basicConfig(
    level='DEBUG',
    format="%(message)s",
)



"""This Module allows for the listing of recipients"""
from libpkpass import LOGGER
from libpkpass.commands.command import Command
from libpkpass.errors import CliArgumentError
from libpkpass.models.cert import Cert
from libpkpass.models.recipient import Recipient

    ####################################################################
class Listrecipients(Command):
    """This class implements the cli functionality to list recipients"""
    ####################################################################
    name = 'listrecipients'
    description = 'List the recipients that pkpass knows about'
    selected_args = Command.selected_args + ['stdin', 'filter']

        ####################################################################
    def _run_command_execution(self):
        """ Run function for class.                                  """
        ####################################################################
        LOGGER.info('Certificate store: %s', self.args["certpath"])
        LOGGER.info('Key store: %s', self.args["keypath"])
        LOGGER.info('CA Bundle file: %s', self.args["cabundle"])
        LOGGER.info('Looking for Key Extension: %s', self.identities.extensions["key"])
        LOGGER.info('Looking for Certificate Extension: %s', self.identities.extensions["certificate"])
        LOGGER.info("Loaded %s identities", len(self.session.query(Recipient).all()))

        if 'filter' in self.args and self.args['filter']:
            identities = self.session.query(Recipient).filter(
                Recipient.name.like(self.args['filter'].replace('*', '%'))
            )
        else:
            identities = self.session.query(Recipient).all()
        for identity in identities:
            yield self._print_identity(
                identity,
                self.session.query(Cert).filter(
                    Cert.recipients.contains(identity)
                ).all()
            )

        ####################################################################
    def _print_identity(self, identity, certs):
        """Print off identity"""
        ####################################################################
        identity_list = []
        identity_list.append(f"{self.color_print(identity.name, 'first_level')}:")
        identity_list.append(f"\t{self.color_print('certs', 'second_level')}:")
        for cert in certs:
            for info in ['verified', 'subject', 'subjecthash', 'issuer', 'issuerhash', 'fingerprint', 'enddate']:
                identity_list.append(f"\t\t{self.color_print(info + ':', 'third_level')} {dict(cert)[info]}")
            identity_list.append('\n')
        return "\n".join(identity_list)

        ####################################################################
    def _validate_args(self):
        """ Ensure arguments are appropriate for this command       """
        ####################################################################
        for argument in ['keypath']:
            if argument not in self.args or self.args[argument] is None:
                raise CliArgumentError(f"'{argument}' is a required argument")

        ####################################################################
    def _validate_identities(self, _=None):
        """ Ensure identities are appropriate for this command       """
        ####################################################################
        return



"""This Module handles the identitydb object"""
from os import path, listdir
from datetime import datetime
from pem import parse_file
from sqlalchemy.orm import sessionmaker
from tqdm import tqdm
from libpkpass.crypto import pk_verify_chain, get_cert_fingerprint, get_cert_subject,\
    get_cert_issuer, get_cert_enddate, get_cert_issuerhash, get_cert_subjecthash
from libpkpass.errors import FileOpenError, CliArgumentError
from libpkpass.models.recipient import Recipient
from libpkpass.models.cert import Cert
from libpkpass.util import create_or_update


    ##########################################################################
class IdentityDB():
    """ User database class.  Contains information about the identities of and
        things pertinent to recipients and their groups and keys.              """
    ##########################################################################

        #######################################################################
    def __init__(self):
        #######################################################################
        self.extensions = {'certificate': ['.cert', '.crt'],
                           'key': '.key'}
        self.cabundle = ""
        self.args = {}

        #######################################################################
    def __repr__(self):
        #######################################################################
        return f"{self.__class__}({self.__dict__})"

        #######################################################################
    def __str__(self):
        #######################################################################
        return f"{self.__dict__}"

        #######################################################################
    def _load_certs_from_external(self, connection_map, usedb=True):
        """Load certificates from external via use of plugin"""
        #######################################################################
        if 'base_directory' in connection_map and connection_map['base_directory']:
            del connection_map['base_directory']
        for key, value in connection_map.items():
            connector = getattr(__import__(key.lower(), fromlist=[key]), key)(value)
            if usedb:
                with open(self.args['db']['path'], 'wb') as db_path:
                    db_path.write(connector.get_db())
            else:
                certs = connector.list_certificates()
                print("Loading certs into database")
                for name, certlist in tqdm(certs.items()):
                    self.load_db(name, certlist)
            break

        #######################################################################
    def _load_from_directory(self, fpath, filetype):
        """ Helper function to read in (keys|certs) and store them correctly """
        #######################################################################
        try:
            print("Loading certs into database")
            for fname in tqdm(listdir(fpath)):
                if fname.endswith(tuple(self.extensions[filetype])):
                    uid = fname.split('.')[0]
                    filepath = path.join(fpath, fname)
                    if filetype == 'certificate':
                        self.load_db(uid, certlist=[x.as_text() for x in parse_file(filepath)])
                    elif filetype == 'key':
                        session = sessionmaker(bind=self.args['db']['engine'])()
                        create_or_update(session,
                                         Recipient,
                                         unique_identifiers=['name'],
                                         **{
                                             'name': uid,
                                             'key': filepath,
                                         },
                                         )
                        session.commit()
        except OSError as error:
            raise FileOpenError(fpath, str(error.strerror)) from error

        #######################################################################
    def load_certs_from_directory(self,
                                  certpath,
                                  connectmap=None,
                                  nocache=False,
                                  usedb=True):
        """ Read in all x509 certificates from directory and name them as found """
        #######################################################################
        session = sessionmaker(bind=self.args['db']['engine'])()
        if not session.query(Recipient).first() or nocache:
            if connectmap:
                self._load_certs_from_external(connectmap, usedb=usedb)
            if certpath:
                self._load_from_directory(certpath, 'certificate')

        #######################################################################
    def load_db(self, identity, certlist=None):
        """ Read in all rsa keys from directory and name them as found
        """
        #######################################################################
        try:
            session = sessionmaker(bind=self.args['db']['engine'])()
            recipient = create_or_update(session, Recipient, dont_update=['key'], **{'name': identity})
            for cert in certlist:
                cert = cert.encode()
                cert_dict = {}
                cert_dict['cert_bytes'] = cert
                cert_dict['verified'] = pk_verify_chain(cert, self.cabundle)
                cert_dict['fingerprint'] = get_cert_fingerprint(cert)
                cert_dict['subject'] = get_cert_subject(cert)
                cert_dict['issuer'] = get_cert_issuer(cert)
                cert_dict['enddate'] = datetime.strptime(get_cert_enddate(cert), '%b %d %H:%M:%S %Y %Z')
                cert_dict['issuerhash'] = get_cert_issuerhash(cert)
                cert_dict['subjecthash'] = get_cert_subjecthash(cert)
                cert = create_or_update(
                    session,
                    Cert,
                    unique_identifiers=['fingerprint'],
                    **cert_dict
                )
                if cert not in recipient.certs:
                    recipient.certs.append(cert)
            session.commit()
        except KeyError as err:
            raise CliArgumentError(
                f"Error: Recipient '{identity}' is not in the recipient database"
            ) from err

    def verify_identity(self, identity):
        session = sessionmaker(bind=self.args['db']['engine'])()
        recipient = session.query(Recipient).filter(Recipient.name == identity).first()
        if not recipient:
            return False
        for cert in session.query(Cert).filter(
                Cert.recipients.contains(recipient)
        ).all():
            if not cert.verified:
                return False
        return True

        #######################################################################
    def load_keys_from_directory(self, fpath):
        """ Read in all rsa keys from directory and name them as found """
        #######################################################################
        if path.isdir(fpath):
            self._load_from_directory(fpath, 'key')
        # We used to print a warning on else here, but it seems unnecessary;
        # if a user is using PIVs/Smartcards, this warning could just annoy
        # them needlessly



"""This Modules allows for distributing created passwords to other users"""
from os import path
from tqdm import tqdm
from libpkpass import LOGGER
from libpkpass.util import dictionary_filter, sort
from libpkpass.commands.command import Command
from libpkpass.password import PasswordEntry
from libpkpass.errors import CliArgumentError
from libpkpass.models.recipient import Recipient

    ####################################################################
class Distribute(Command):
    """This Class implements the CLI functionality for ditribution"""
    ####################################################################
    name = 'distribute'
    description = 'Distribute existing password entry/ies to another entity [matching uses python fnmatch]'
    selected_args = Command.selected_args + ['pwname', 'pwstore', 'users', 'groups', 'stdin',
                                             'min_escrow', 'escrow_users', 'keypath', 'nopassphrase',
                                             'nosign', 'card_slot', 'noescrow']

        ####################################################################
    def _run_command_execution(self):
        """ Run function for class.                                      """
        ####################################################################
        yield from self._confirm_pdb()
        self.recipient_list.append(str(self.args['identity']))
        self.recipient_list = list(set(self.recipient_list))
        yield from self._confirm_recipients()
        for dist_pass, _ in tqdm(self.filtered_pdb.items()):
            password = PasswordEntry()
            password.read_password_data(dist_pass)
            if self.args['identity'] in password.recipients.keys():
                # we shouldn't modify escrow on distribute
                self.args['min_escrow'] = None
                self.args['escrow_users'] = None
                plaintext_pw = password.decrypt_entry(
                    self.identity,
                    passphrase=self.passphrase,
                    card_slot=self.args['card_slot']
                )
                password.add_recipients(
                    secret=plaintext_pw,
                    distributor=self.args['identity'],
                    recipients=self.recipient_list,
                    session=self.session,
                    passphrase=self.passphrase,
                    card_slot=self.args['card_slot'],
                    pwstore=self.args['pwstore']
                )

                password.write_password_data(dist_pass)

        ####################################################################
    def _validate_args(self):
        ####################################################################
        for argument in ['pwname', 'keypath']:
            if argument not in self.args or self.args[argument] is None:
                raise CliArgumentError(f"'{argument}' is a required argument")

        ####################################################################
    def _confirm_pdb(self):
        ####################################################################
        self.filtered_pdb = dictionary_filter(
            path.join(self.args['pwstore'], self.args['pwname']),
            self.passworddb.pwdb,
            [self.args['identity'], 'recipients']
        )
        yield "The following password files have matched:"
        yield "\n".join(self.filtered_pdb.keys())
        correct = input("Are these correct? (y/N) ")
        if not correct or correct.lower()[0] == 'n':
            self.args['pwname'] = input("Please try a new filter: ")
            yield from self._confirm_pdb()

        ####################################################################
    def _confirm_recipients(self):
        ####################################################################
        not_in_db = []
        in_db = [x.name for x in self.session.query(Recipient).all()]
        for recipient in self.recipient_list:
            if recipient not in in_db:
                not_in_db.append(recipient)
        if not_in_db:
            LOGGER.warning(
                "The following recipients are not in the db, removing %s", ', '.join(not_in_db)
            )
            self.recipient_list = [x for x in self.recipient_list if x not in not_in_db]
        yield "The following users will receive the password: "
        yield ", ".join(sort(self.recipient_list))
        correct = input("Are these correct? (y/N) ")
        if not correct or correct.lower()[0] == 'n':
            self.recipient_list = input("Please enter a comma delimited list: ")
            self.recipient_list = list({x.strip() for x in self.recipient_list.split(",")})
            yield from self._confirm_recipients()



from sqlalchemy import String
from sqlalchemy.orm import relationship
from sqlalchemy.schema import Column, ForeignKey, Table
from libpkpass.models import Base

recipient_cert = Table(
    'recipient_cert', Base.metadata,
    Column('recipient_name', String(), ForeignKey('recipient.name')),
    Column('cert_fingerprint', String(), ForeignKey('cert.fingerprint')),
)

class Recipient(Base):
    """table of health status of services"""
    __tablename__ = 'recipient'
    name = Column(String(), primary_key=True)
    key = Column(String(), nullable=True)
    certs = relationship("Cert", secondary=recipient_cert, backref='recipients', lazy='joined')

        #########################################
    def __repr__(self):
        #########################################
        return f"<Recipient {self.name}>"

    def __iter__(self):
        for c in self.__table__.columns:
            yield c.name, getattr(self, c.name)



#!/usr/bin/env python3
"""This Module tests iddb module"""
import unittest
import os.path
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from libpkpass.identities import IdentityDB
from libpkpass.models.recipient import Recipient

class TestBasicFunction(unittest.TestCase):
    """This class tests the iddb class"""
    def setUp(self):
        self.certdir = 'test/pki/intermediate/certs'
        self.keydir = 'test/pki/intermediate/private'
        self.cabundle = 'test/pki/intermediate/certs/ca-bundle'
        db_path = 'test/pki/intermediate/certs/rd.db'
        self.idobj = {
            'args': {
                'db': {
                    'uri': f"sqlite+pysqlite:///{db_path}",
                    'engine': create_engine(f"sqlite+pysqlite:///{db_path}")
                }
            }
        }
        self.session = sessionmaker(bind=self.idobj['args']['db']['engine'])()

    def test_certificate_loading(self):
        """Load a cert into our test"""
        idobj = IdentityDB()
        idobj.args = self.idobj['args']
        idobj.load_certs_from_directory(self.certdir, self.cabundle)
        for identity in ('r1', 'r2', 'r3'):
            assert identity in [x[0] for x in self.session.query(Recipient.name).all()]

    def test_key_loading(self):
        """Load a key into our test"""
        idobj = IdentityDB()
        idobj.args = self.idobj['args']
        idobj.load_keys_from_directory(self.keydir)
        for identity in ('r1', 'r2', 'r3'):
            assert identity in [x[0] for x in self.session.query(Recipient.name).all()]
            test_id = self.session.query(Recipient).filter(Recipient.name == identity).first()
            assert os.path.isfile(test_id.key)

    def test_keycert_loading(self):
        """Load a keycert into our test"""
        idobj = IdentityDB()
        idobj.args = self.idobj['args']
        idobj.load_certs_from_directory(self.certdir, self.cabundle)
        idobj.load_keys_from_directory(self.keydir)
        for identity in ('r1', 'r2', 'r3'):
            assert identity in [x[0] for x in self.session.query(Recipient.name).all()]
            test_id = self.session.query(Recipient).filter(Recipient.name == identity).first()
            assert os.path.isfile(test_id.key)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""This module tests the distribute module"""
import builtins
import unittest
import mock
from libpkpass.commands.cli import Cli
from libpkpass.errors import CliArgumentError, DecryptionError
from .basetest.basetest import patch_args, ERROR_MSGS

class DistributeTests(unittest.TestCase):
    """This class tests the distribute class"""
    def test_recipient_not_in_database(self):
        """Test what happens with a recipient is not within the appropriate directory"""
        with self.assertRaises(CliArgumentError) as context:
            with patch_args(subparser_name='distribute', identity='bleh', nopassphrase='true',
                            pwname='test'):
                Cli().run()
        self.assertEqual(
            context.exception.msg,
            ERROR_MSGS['rep']
        )

    def test_distribute_cli_error(self):
        """Test what happens with pwname is not supplied"""
        with self.assertRaises(CliArgumentError) as context:
            with patch_args(subparser_name='distribute', identity='r1', nopassphrase='true',
                            pwname=None):
                Cli().run()
        self.assertEqual(context.exception.msg, ERROR_MSGS['pwname'])

    def test_distribute_success(self):
        """Test a successful distribute"""
        ret = True
        try:
            with patch_args(subparser_name='distribute', identity='r1', nopassphrase='true',
                            pwname='test'):
                with mock.patch.object(builtins, 'input', lambda _: 'y'):
                    Cli().run()
        except DecryptionError:
            ret = False
        self.assertTrue(ret)


if __name__ == '__main__':
    unittest.main()



"""This module allows for inspecting metadata passwords"""
from os import path, linesep
from datetime import datetime
from libpkpass.password import PasswordEntry
from libpkpass.commands.command import Command
from libpkpass.errors import CliArgumentError

    ####################################################################
class Info(Command):
    """This class implements the display of metadata to users"""
    ####################################################################
    name = 'info'
    description = 'Displays metadata about a password'
    selected_args = Command.selected_args + ['pwname', 'pwstore']

        ####################################################################
    def _run_command_execution(self):
        """ Run function for class.                                      """
        ####################################################################
        password = PasswordEntry()
        password.read_password_data(path.join(self.args['pwstore'], self.args['pwname']))

        # Metadata
        yield self.color_print("Metadata:", "first_level")
        for key, value in password.metadata.items():
            yield f"  {self.color_print(key.capitalize(), 'second_level')}: {value}"

        # Escrow
        if password.escrow:
            yield self.color_print("\nEscrow Groups:", "first_level")
            for group_key, group_value in password.escrow.items():
                yield f"  {self.color_print(group_key, 'second_level')}:"
                for key, value in group_value['metadata'].items():
                    yield f"    {self.color_print(key + ':', 'third_level')} {str(value)}"

                yield f"    {self.color_print('Share Holders:', 'third_level')} {', '.join(list(group_value['recipients'].keys()))}"
                yield f"    {self.color_print('Total Group Share Holders:', 'third_level')} {len(list(group_value['recipients'].keys()))}"

                timestamp_list = [x['timestamp'] for x in list(group_value['recipients'].values()) if 'timestamp' in x]
                if timestamp_list:
                    timestamp = int(round(float(min(timestamp_list))))
                    timestamp = datetime.fromtimestamp(timestamp).strftime('%Y-%m-%d %H:%M:%S')
                    yield f"    {self.color_print('Group creation time:', 'third_level')} {timestamp}"

        # Recipients
        yield f"{self.color_print(linesep + 'Recipients:', 'first_level')} {', '.join(list(password.recipients.keys()))}"
        yield f"{self.color_print('Total Recipients:', 'first_level')} {len(list(password.recipients.keys()))}"
        rec_timestamp = int(round(float(min([x['timestamp'] for x in list(password.recipients.values()) if 'timestamp' in x]))))
        rec_timestamp = datetime.fromtimestamp(rec_timestamp).strftime('%Y-%m-%d %H:%M:%S')
        yield f"{self.color_print('Earliest distribute timestamp:', 'first_level')} {rec_timestamp}"

        ####################################################################
    def _validate_args(self):
        ####################################################################
        for argument in ['pwname', 'keypath']:
            if argument not in self.args or self.args[argument] is None:
                raise CliArgumentError(f"'{argument}' is a required argument")



#!/usr/bin/env python3
"""This module tests the update module"""
import getpass
import builtins
import unittest
import mock
from libpkpass.commands.cli import Cli
from libpkpass.errors import DecryptionError
from .basetest.basetest import patch_args

class UpdateTests(unittest.TestCase):
    """This class tests the update class"""
    def test_update_success(self):
        """Test a successful update of a password"""
        ret = True
        try:
            with patch_args(subparser_name='update', identity='r3', nopassphrase='true',
                            pwname='gentest', no_cache=True):
                with mock.patch.object(builtins, 'input', lambda _: 'y'):
                    with mock.patch.object(getpass, 'getpass', lambda _: 'y'):
                        Cli().run()
        except DecryptionError:
            ret = False
        self.assertTrue(ret)



#!/usr/bin/env python3
"""This is a super class to handle commonalities between the cli and interpreter"""
from argparse import ArgumentParser
from os import path
from libpkpass.util import set_default_subparser
from libpkpass.commands import card, clip, create, delete, \
    distribute, export, generate, info, listrecipients, \
    modify, recover, rename, show, populate, update, verifyinstall
import libpkpass.commands.fileimport as pkimport
import libpkpass.commands.list as pklist

    ##############################################################################
class PkInterface():
    """ Class for parsing command line.  Observes subclasses of Command to Register
    those commands in the actions list.                                        """
    ##############################################################################
        ####################################################################
    def __init__(self):
        """ Intialization function for class. Register all subcommands   """
        ####################################################################
        # Hash of registered subparser actions, mapping string to actual subparser
        self.actions = {}
        home = path.expanduser("~")
        defaultrc = None
        for file in ['.pkpassrc', '.pkpassrc.yml', '.pkpassrc.yaml']:
            if path.isfile(path.join(home, file)):
                defaultrc = path.join(home, file)
        if not defaultrc:
            defaultrc = path.join(home, '.pkpassrc')
        self.parser = ArgumentParser(
            description='Public Key Password Manager')
        self.parser.set_default_subparser = set_default_subparser
        self.parser.add_argument(
            '--config', type=str,
            help="Path to a PKPass configuration file.  Defaults to '~/.pkpassrc{,.yml,.yaml}'",
            default=defaultrc
        )
        self.parser.add_argument('--debug', action='store_true',
                                 help="Errors are more verbose")
        self.subparsers = self.parser.add_subparsers(
            help='sub-commands', dest='subparser_name')

        card.Card(self)
        clip.Clip(self)
        create.Create(self)
        delete.Delete(self)
        distribute.Distribute(self)
        export.Export(self)
        generate.Generate(self)
        pkimport.Import(self)
        info.Info(self)
        pklist.List(self)
        listrecipients.Listrecipients(self)
        modify.Modify(self)
        recover.Recover(self)
        rename.Rename(self)
        show.Show(self)
        populate.Populate(self)
        update.Update(self)
        verifyinstall.VerifyInstall(self)

        ####################################################################
    def register(self, command_obj, command_name, command_description):
        """ Register command objects and names using an observer pattern """
        ####################################################################
        self.actions[command_name] = command_obj
        parser = self.subparsers.add_parser(
            command_name, help=command_description)
        command_obj.register(parser)



#!/usr/bin/env python3
"""This Module implements a CLI"""
from libpkpass.util import show_version
from libpkpass.commands.pkinterface import PkInterface
from libpkpass.commands.interpreter import Interpreter

    ##############################################################################
class Cli(PkInterface):
    """ Class for parsing command line.  Observes subclasses of Command to Register
    those commands in the actions list.                                        """
    ##############################################################################

        ####################################################################
    def __init__(self):
        """ Intialization function for class. Register all subcommands   """
        ####################################################################
        PkInterface.__init__(self)

        Interpreter(self)
        self.parser.add_argument(
            '--version', action='store_true', help='Show the version of PkPass and exit')

        self.parser.set_default_subparser(self.parser, name='interpreter')
        self.parsedargs = self.parser.parse_args()

    def run(self):
        if 'version' in self.parsedargs and self.parsedargs.version:
            return show_version()
        return self.actions[self.parsedargs.subparser_name].run(self.parsedargs)



#!/usr/bin/env python3
"""This module tests the generate module"""
import getpass
import builtins
import unittest
import mock
from libpkpass.commands.cli import Cli
from libpkpass.errors import CliArgumentError, DecryptionError
from .basetest.basetest import patch_args, ERROR_MSGS

class GenerateTests(unittest.TestCase):
    """This class tests the generate class"""
    def test_generate_success(self):
        """Test a successful generation of a password"""
        ret = True
        try:
            with patch_args(subparser_name='generate', identity='r3', nopassphrase='true',
                            pwname='gentest'):
                with mock.patch.object(builtins, 'input', lambda _: 'y'):
                    with mock.patch.object(getpass, 'getpass', lambda _: 'y'):
                        Cli().run()
        except DecryptionError:
            ret = False
        self.assertTrue(ret)

    def test_generate_no_pass(self):
        """test what happens with pwname is not supplied"""
        with self.assertRaises(CliArgumentError) as context:
            with patch_args(subparser_name='generate', identity='r3', nopassphrase='true',
                            pwname=None):
                with mock.patch.object(builtins, 'input', lambda _: 'y'):
                    with mock.patch.object(getpass, 'getpass', lambda _: 'y'):
                        Cli().run()
        self.assertEqual(context.exception.msg, ERROR_MSGS['pwname'])

if __name__ == '__main__':
    unittest.main()



"""This module allows for the population of external password stores"""
from os import path, remove
from re import finditer
from base64 import standard_b64encode
from ruamel.yaml import YAML
from yaml import dump
from libpkpass import LOGGER
from libpkpass.commands.show import Show
from libpkpass.crypto import puppet_password
from libpkpass.password import PasswordEntry
from libpkpass.errors import CliArgumentError
from libpkpass.models.recipient import Recipient

    ####################################################################
class Populate(Show):
    """This class implements the CLI functionality of password store integration"""
    ####################################################################
    name = 'populate'
    description = 'Populate external resource with password, currently supports: puppet_eyaml'
    selected_args = ['pwname', 'cabundle', 'card_slot', 'nopassphrase',
                     'all', 'certpath', 'color', 'identity', 'keypath',
                     'nocache', 'noverify', 'pwstore', 'quiet',
                     'theme_map', 'type', 'value', 'verbosity']

        ####################################################################
    def _run_command_execution(self):
        """ Run function for class.                                      """
        ####################################################################
        # currently only support puppet
        pop_type = self.args['type']
        if pop_type not in ['puppet_eyaml', 'kubernetes']:
            raise CliArgumentError(f"'{pop_type}' is an unsupported population type")

        if pop_type == 'kubernetes':
            yield from self._handle_kubernetes()
        elif self.args['all']:
            for password in self.args['populate'][pop_type]['passwords'].keys():
                pwentry = PasswordEntry()
                self._decrypt_wrapper(
                    self.args['pwstore'],
                    pwentry,
                    password,
                    pop_type
                )

        elif self.args['pwname'] not in self.args['populate'][pop_type]['passwords'].keys():
            raise CliArgumentError(f"'{self.args['pwname']}' doesn't have a mapping in {pop_type}")
        else:
            password = PasswordEntry()
            yield from self._decrypt_wrapper(
                self.args['pwstore'],
                password,
                self.args['pwname'],
                pop_type
            )

        ####################################################################
    def _handle_kubernetes(self):
        """This function creates a file containing kube secrets"""
        ####################################################################
        k_conf = self.args['populate']['kubernetes']
        plaintext = {}
        for password in k_conf['needed']:
            pwentry = PasswordEntry()
            pwentry.read_password_data(path.join(self.args['pwstore'], password))
            plaintext_pw = self._decrypt_password_entry(pwentry)
            plaintext[password] = plaintext_pw
        if 'output' in k_conf:
            if self.args['pwstore'] in k_conf['output']:
                raise CliArgumentError("Kubernetes output file should not exist in password store")
            if path.isfile(k_conf['output']):
                remove(k_conf['output'])
        for args in k_conf['passwords']:
            data = self._kubernetes_match_loop(args, plaintext)
            kube_map = {
                'kind': 'Secret',
                'apiVersion': args['apiVersion'],
                'metadata': args['metadata'],
                'data': data,
                'type': args['type']
            }
            if self.args['value'] or 'output' not in k_conf:
                yield f"---\n{dump(kube_map)}"
            else:
                with open(k_conf['output'], 'a', encoding='ASCII') as fname:
                    print(f"---\n{dump(kube_map)}", file=fname)

        ####################################################################
    def _kubernetes_match_loop(self, args, plaintext):
        """This matches ${example} inside the arg defs and generates a
        dictionary of b64 encoded values with the actual password values
        placed instead of ${}"""
        ####################################################################
        data = {}
        for key, value in args['data'].items():
            new_pw = str(value)
            match_iter = finditer(r'\$\{([^${]*)\}', new_pw)
            if not match_iter:
                data[key] = standard_b64encode(new_pw.encode('ASCII')).decode()
            else:
                for match_values in match_iter:
                    match_value = match_values.group(1)
                    new_pw = new_pw.replace(f"%{{{match_value}}}", plaintext[match_value])
                data[key] = standard_b64encode("".join(new_pw).encode('ASCII')).decode()
        return data

        ####################################################################
    def _handle_puppet(self, plaintext_pw, pwname):
        ####################################################################
        directory = path.expanduser(self.args['populate']['puppet_eyaml']['directory'])
        if not path.isdir(directory):
            raise CliArgumentError(f"'{directory}' is not a directory")
        puppet_bin = path.expanduser(self.args['populate']['puppet_eyaml']['bin'])
        if not path.isfile(puppet_bin):
            raise CliArgumentError(f"'{puppet_bin}' is not a file")
        pkcs7_pass = puppet_password(puppet_bin, plaintext_pw)
        if self.args['value']:
            yield pkcs7_pass
        else:
            for hiera_file, names in self.args['populate']['puppet_eyaml']['passwords'][pwname].items():
                with open(path.join(directory, hiera_file), 'r', encoding='ASCII') as data_file:
                    yield f"Updating: {hiera_file}"
                    yaml = YAML()
                    yaml.indent(mapping=2, sequence=4, offset=2)
                    yaml.preserve_quotes = True
                    hiera_yaml = yaml.load(data_file.read())
                    for name in names:
                        yield f"Updating: {name}"
                        hiera_yaml[name] = pkcs7_pass
                with open(path.join(directory, hiera_file), 'w', encoding='ASCII') as data_file:
                    yaml.dump(hiera_yaml, data_file)

        ####################################################################
    def _validate_args(self):
        ####################################################################
        for argument in ['keypath', 'type']:
            if argument not in self.args or self.args[argument] is None:
                raise CliArgumentError(f"'{argument}' is a required argument")

        ####################################################################
    def _decrypt_wrapper(self, directory, password, pwname, pop_type):
        """Decide whether to decrypt normally or for escrow"""
        ####################################################################
        password.read_password_data(path.join(directory, pwname))
        plaintext_pw = self._decrypt_password_entry(password)
        if pop_type == 'puppet_eyaml':
            yield from self._handle_puppet(plaintext_pw, pwname)

        ####################################################################
    def _decrypt_password_entry(self, password):
        """This decrypts a given password entry"""
        ####################################################################
        plaintext_pw = password.decrypt_entry(
            identity=self.identity, passphrase=self.passphrase, card_slot=self.args['card_slot']
        )
        distributor = password.recipients[self.identity['name']]['distributor']
        if not self.args['noverify']:
            result = password.verify_entry(
                self.identity['uid'],
                self.identities,
                distributor,
                self.session.query(Recipient).filter(Recipient.name==distributor).first().certs,
            )
            if not result['sigOK']:
                LOGGER.warning(
                    "Could not verify that %s correctly signed your password entry.",
                    result['distributor']
                )
            if not result['certOK']:
                LOGGER.warning(
                    "Could not verify the certificate authenticity of user '%s'.",
                    result['distributor']
                )
        return plaintext_pw
